#                                                         -*- Perl -*-
# jargon-fpw - FreePWING script for The Jargon File
# $Id: wdic-fpw.pl 1.1.0.6 2000/07/16 22:53:56 satomii Exp $
#
# Warning: This script has been tested only with ActivePerl 5.6.0 Build 613
# on Win32 environment. I don't and can't care about Unix environment, sorry.
#
# Copyright (C) 2000, Rei <rei@wdic.org>.
# This program is distributed in the hope that it will be useful, but
# without any warranty. See the GNU General Public License for the details.
#
# This script is not so neat as you see - that's because I am not a skilled
# Perl programmer. Please re-write if you don't like it :-)
#
require 5.005;

use Getopt::Long;
use FreePWING::FPWUtils::FPWParser;
use Jcode;

use strict "vars";

my $wdic_title = 'ʐMp̊bm';

#
# regular expressions for SJIS.
#
my $wdic_midpoint = '(?:\x81\x45)';					# E
my $wdic_column = '(?:\x81\x06)';					# 
my $wdic_reference = '(?:\x81[\x9a\x9f\xa8])';		# 
my $wdic_uri = '(?:\x81[\xa4\xa5])';				# 
my $wdic_plugin = '(?:\x81\x97)';					# 

my $sjis_any = '(?:[\x81-\x9f\xe0-\xfc][\x40-\x7e\x80-\xfc])';
my $sjis_hankaku = '[\xa1-\xdf]';

my $ascii_char = '[\x21-\x7f]';						# not include spaces.
my $ascii_any = '(?:[\x20-\x7f]|$sjis_hankaku)';	# include ' '.

my $char_any = '(?:$ascii_any|$sjis_any)';

#
# base names and titles of the input files (these string can be changed
# into other kanji-code).
#
my %wdic_name = (
	COMM => 'ʐM',
	DEV => 'u',
	ELEC => 'dq',
	GAME => 'Q[',
	INFO => '',
	KNOW => 'dZm',
	MISC => 'Gwm',
	MOE1 => 'G 1',
	MOE2 => 'G 2',
	ORG => 'c̕',
	PASO => 'p 2',
	RAIL => 'S',
	RCMP => 'SЕ',
	SOFT => 'i',
	STD => 'Ki',
	TECH => 'Zp',
	TEL => 'db',
	TRAN => 'ʐMZp',
	YOGO => 'p 1'
);

#
# misc variables.
#
my($srcdir) = '.';
my($fpwtext, $fpwheading, $fpwword2, $fpwmenu);
my($opt_usage, $opt_charset, $opt_bullet, $opt_sort, $opt_gaiji);
my(%wdic_index);

#-----------------------------------------------------------------------#
#	main routine														#
#-----------------------------------------------------------------------#

my($sort_file, $dic, $lines);

#
# initialization.
#
&initialize();
if (!(-f "$srcdir/WDIC.MAN")) {
	&fatal("Missing required file: $srcdir/WDIC.MAN");
}

initialize_fpwparser('text' => \$fpwtext,
		'heading' => \$fpwheading,
		'word2' => \$fpwword2,
		'menu' => \$fpwmenu);

#
# read the entries and write them to the temporary file. this is done to
# sort the entries throughout the separated dictionaries.
#
$sort_file = "$work_directory/wdic.tmp";

open(OUT, "| $opt_sort > $sort_file") ||
	&fatal("Cannot create working file \'$sort_file' ($!)");
close(OUT);
if ($?) {
	unlink($sort_file);
	&fatal("Failed to invoke '$opt_sort' command");
}

open(OUT, "| $opt_sort > $sort_file") ||
	&fatal("Cannot create working file \'$sort_file' ($!)");
autoflush OUT;
foreach $dic (keys %wdic_name) {
	$lines += &wdic_read($dic);
}
close(OUT);
if ($? || !$lines) {
	unlink($sort_file);
	&fatal("No entry was found, abort");
}

#
# reopen the sorted file and convert via FreePWING.
#
open(IN, "<$sort_file") ||
	&fatal("Cannot open working file \'$sort_file' ($!)");
binmode(IN);
print "Converting entries...\n";

while (<IN>) {
	s/\x0d\x0a$//;
	next if (/^$/);
	&convert_entry(split(/\t/));
}

close(IN);
unlink($sort_file);

#
# write the plug-in and menu entries.
#
&wdic_ext();
&wdic_menu();

#
# clean up.
#
finalize_fpwparser('text' => \$fpwtext,
		'heading' => \$fpwheading,
		'word2' => \$fpwword2,
		'menu' => \$fpwmenu);
exit(0);

#-----------------------------------------------------------------------#
#	initialization														#
#-----------------------------------------------------------------------#
#
# void initialize(void)
#
sub initialize
{
	#
	# look up the command line options.
	#
	if (!GetOptions('help|h|?' => \$opt_usage,
			'charset|c:s' => \$opt_charset,
			'bullet|b:s' => \$opt_bullet,
			'sort|s:s' => \$opt_sort,
			'gaiji|g' => \$opt_gaiji)
			|| 1 < scalar(@ARGV)) {
		usage(1);
	}
	usage(0) if ($opt_usage);

	#
	# if charset is not specified, look up the LANG environment variable.
	#
	if ((!defined($opt_charset) || '' eq $opt_charset)
			&& $ENV{'LANG'} =~ /^ja_JP\.([a-zA-Z]+)$/) {
		$opt_charset = $1;
	}

	#
	# ensure that the charset can be used with the Jcode module.
	#
	$opt_charset =~ tr/A-Z/a-z/;
	if ("ujis" eq $opt_charset
			|| "eucjp" eq $opt_charset) {
		$opt_charset = "euc"
	}
	if (!defined($opt_charset) || '' eq $opt_charset) {
		if ($ENV{'OS'} =~ /^Windows/
				|| $ENV{'COMSPEC'} =~ /COMMAND\.COM$/i
				|| $ENV{'COMSPEC'} =~ /CMD\.EXE$/i) {
			# Windows or MS-DOS environment.
			$opt_charset = "sjis";
		} else {
			# default to euc for non-Windows system.
			$opt_charset = "euc";
		}
	} elsif ('sjis' ne $opt_charset
			&& 'euc' ne $opt_charset
			&& 'jis' ne $opt_charset
			&& 'iso_2022_jp' ne $opt_charset) {
		# unknown charset.
		&fatal("Cannot recognize output charset: " .
				"$opt_charset");
	}
	print "Detected display charset: $opt_charset\n";

	#
	# ensure that other option values are valid.
	#
	$opt_bullet = '' if (!defined($opt_bullet) || '' eq $opt_bullet);
	$opt_bullet = Jcode::convert($opt_bullet, 'euc');

	if (!defined($opt_sort) || '' eq $opt_sort) {
		if (defined($ENV{'SORT'}) && '' ne $ENV{'SORT'}) {
			$opt_sort = $ENV{'SORT'};
		} elsif ($ENV{'OS'} =~ /^Windows/
				|| $ENV{'COMSPEC'} =~ /COMMAND\.COM$/i
				|| $ENV{'COMSPEC'} =~ /CMD\.EXE$/i) {
			# Windows or MS-DOS environment.
			$opt_sort = 'sort.exe /REC 65535';
		} else {
			$opt_sort = '/usr/bin/sort';
		}
	}

	#
	# fix input directory name.
	#
	if (defined($ARGV[0])) {
		$srcdir = shift(@ARGV);
		$srcdir =~ s/^(.*)[\\\/]$/\1/;
	}
}


#-----------------------------------------------------------------------#
#	preprocess															#
#-----------------------------------------------------------------------#
#
# int wdic_read(LPCSTR dic);
# read each input file and write the entries into a temporary file.
# the output will be sorted in order of the japanese syllabry (or
# alphabetically).
#
sub wdic_read
{
	my($dic, $line, $index, $count);

	$dic = $_[0];
	return unless (defined($wdic_name{$_[0]}));
	$wdic_name{$dic} = Jcode::convert($wdic_name{$dic}, 'euc');

	$dic = "$srcdir/$dic.DIC";
	if (!(-f $dic)) {
		&warn("Dictionary \'$dic\' not found, skip");
		return(0);
	}

	open(IN, "<$dic") || &fatal("Cannot open \'$dic\' ($!).");
	binmode(IN);
	print("Reading \'$dic\'...\n");

	while (<IN>) {
		last if (!defined($_));
		s/\x0d\x0a$//;
		$line++;

		while (/^$/) {
			$_ = <IN>;
			s/\x0d\x0a$//;
			$line++;
			last if (!defined($_) || /^[%\#]/);
		}
		last if (!defined($_));

		if (/^$/) {
			#
			# end of the current block.
			#
			$index = '';

		} elsif (s/^%//) {
			#
			# a line for an index character.
			#
			if (/L/) {
				$index = "0000";
			} elsif (//) {
				$index = "0001";
			} else {
				$index = sprintf("%02x%02x", unpack('CC', $_));
			}
			#&jprint("$_ ");

		} elsif (s/^\#//) {
			#
			# start of a new word.
			#
			my($word) = $_;

			# store this word into the index array.
			push(@{$wdic_index{$_[0]}}, $word);

			$count++;
			print OUT "\n$index\t";

			# read next line and extract the reading.
			$_ = <IN>;
			$line++;
			s/\x0d\x0a$//;
			if (!s/^ {8}//) {
				&fatal("Invalid dictionary format " .
						"in \'$dic\' at line $line: $_");
			}

			if ($index =~ /^00/ || $index =~ /^82[67]/) {
				print OUT make_sortword($word);
			} elsif (/\[((?:$sjis_any|[^\]])+)\]/) {
				print OUT make_sortword($1);
			} else {
				print OUT make_sortword($word);
			}
			print OUT "\t$_[0]\t$word\t$_";

		} elsif (s/^ {8}// && '' ne $index) {
			#
			# a description line contiuned from the previous one.
			#
			print OUT "\t$_";
		}
	}

	close(IN);
	print OUT "\n";
	print "  $count words found.\n";
	return($count);
}


#-----------------------------------------------------------------------#
#	conversion															#
#-----------------------------------------------------------------------#
#
# void convert_entry(LPCSTR *lines);
#
sub convert_entry
{
	my($head, $read, $dic, $word, @lines) = @_;
	my($text_pos, $heading_pos, $add_break, $s);
	my($index, @keys);

	$index = &get_tag($word, $dic);
	if ('' eq $index) {
		&fatal("Cannot determine tag for \'$word\' (in $dic)");
	}
	$index =~ s/\t.*$//;
	&jprint("$index: $word");

	$fpwtext->new_entry() ||
		&fatal("\nCannot start new entry", $fpwtext);
	$fpwheading->new_entry() ||
		&fatal("\nCannot start new entry", $fpwheading);

	$text_pos = $fpwtext->entry_position();
	$heading_pos = $fpwheading->entry_position();
	$fpwtext->add_tag($index) ||
		&fatal("\nFailed to add an entry tag", $fpwtext);

	#
	# write a tag, heading and keyword.
	#
	$word =~ s/(\s+\([0-9]+\))$//;
	$index = $1;
	$s = Jcode::convert($word, 'euc', 'sjis');
	$fpwheading->add_text($s) ||
		&fatal("\nCannot add heading", $fpwheading);

	if (!$fpwtext->add_keyword_start()
			|| !$fpwtext->add_text($s)
			|| !$fpwtext->add_keyword_end()) {
		&fatal("\nCannot add heading text", $fpwtext);
	}
	if ('' ne $index) {
		$fpwtext->add_text($index) ||
			&fatal("\nFailed to add text", $fpwtext);
	}
	if (!$fpwtext->add_text(" : ")
			|| !$fpwtext->add_text($wdic_name{$dic})
			|| !$fpwtext->add_newline()) {
		&fatal("\nFailed to append dictionary name", $fpwtext);
	}

	#
	# write the information line(s) for this word.
	#
	$_ = shift(@lines);
	while (0 < scalar(@lines)
			&& $lines[0] !~
			/^$wdic_midpoint|$wdic_reference|$wdic_plugin|$wdic_uri/) {
		s/([,\.])$/\1 /;
		$_ .= " " if (!/$char_any $/ && $lines[0] =~ /^$ascii_char/);
		$_ .= shift(@lines);
	}
	&add_line(0, $_);

	#
	# register search words.
	#
	$fpwword2->add_entry($s,
			$heading_pos, $heading_file_name,
			$text_pos, $text_file_name) ||
		&fatal("\nCannot add search word", $fpwword2);
	print " (";

	$read =~ s/^[0-9]+//;
	if (/\[((?:$sjis_any|[^\]])+)\]/) {
		push(@keys, $1) if ($1 ne $word && '???' ne $1);
	}
	if (/\(((?:$sjis_any|[^\)])+)\)/ && $1 ne $word) {
		$_ = $1;
		s/\[(?:$sjis_any|[^\]])+\]//g;
		push(@keys, split(/[:;\/] +/));
	}

	$s = '';
	foreach $_ (@keys) {
		if (!/^$/) {
			s/^\s+//;
			s/\s+$//;
			$_ = &ensure_str($_);
			if (!/^$ascii_char+$/) {
				$_ = Jcode::convert($_, 'euc', 'sjis');
			}
			$fpwword2->add_entry($_,
					$heading_pos, $heading_file_name,
					$text_pos, $text_file_name) ||
				&fatal(")\nCannot add search word: $_", $fpwword2);
			&jprint("$s$_");
			$s = ', ' if ('' eq $s);
		}
	}
	print ")\n";

	#
	# write the body text.
	#
	$add_break = 0;
	while (0 < scalar(@lines)
			&& $lines[0] =~ s/^$wdic_midpoint//) {
		$_ = shift(@lines);
		next if (/^$/);
		while (0 < scalar(@lines) && $lines[0] =~ s/^  ([^ ])/\1/) {
			s/([,\.])$/\1 /;
			$_ .= shift(@lines);
		}
		$fpwtext->add_text($opt_bullet) ||
			&fatal("Failed to write a bullet", $fpwtext);
		&add_line(0, $_);
		$add_break = 1;
	}

	#
	# write examples and/or columns, specifying no-wrapping attribute.
	#
	if (0 < $add_break) {
		$fpwtext->add_newline();
		$add_break = 0;
	}
	$fpwtext->add_indent_level(2);
	while (defined($lines[0]) &&
			$lines[0] !~ /^($wdic_reference|$wdic_uri|$wdic_plugin)/) {
		$_ = shift(@lines);
		next if (/^$/);
		&add_line(1, $_);
		$add_break = 1;
	}

	#
	# write references, related URIs and/or plugin filenames.
	#
	if (0 < $add_break) {
		$fpwtext->add_newline();
		$add_break = 0;
	}

	while (0 < scalar(@lines)) {
		$_ = shift(@lines);
		#chomp();
		next if (/^$/);

		while (0 < scalar(@lines) && $lines[0] =~ s/^ {2}//) {
			#
			# continued from the previous line.
			#
			#chomp($lines[0]);
			$lines[0] =~ s/^\(/ (/;
			$_ .= shift(@lines);
		}

		if (s/^($wdic_reference)(.+)$/\2/) {
			#
			# reference to another WDIC entry.
			#
			$s = Jcode::convert($1, 'euc', 'sjis');
			$fpwtext->add_text($s) ||
				&fatal("Failed to add text", $fpwtext);
			($index, $s) = split(/\t/, &get_tag($_));
			$_ = Jcode::convert($_, 'euc', 'sjis');
			if ('' eq $index) {
				$fpwtext->add_text($_) ||
					&fatal("Cannot add text", $fpwtext);
			} elsif (!$fpwtext->add_reference_start()
					|| !$fpwtext->add_text($_)
					|| !$fpwtext->add_reference_end($index)
					|| !$fpwtext->add_text(' : ')
					|| !$fpwtext->add_text($wdic_name{$s})) {
				&fatal("Failed to add a reference", $fpwtext);
			}

		} elsif (/^$wdic_plugin/) {
			#
			# plugin file - multimedia data not supported yet.
			#
			$fpwtext->add_text(Jcode::convert($_, 'euc', 'sjis')) ||
				&fatal("Failed to add text", $fpwtext);

		} else {
			#
			# maybe an URI.
			#
			$fpwtext->add_text(Jcode::convert($_, 'euc', 'sjis')) ||
				&fatal("Failed to add text", $fpwtext);
		}

		$fpwtext->add_newline();
		$add_break = 1;
	}

	$fpwtext->add_newline() if (0 < $add_break);
}


#-----------------------------------------------------------------------#
#	convertion subroutines												#
#-----------------------------------------------------------------------#
#
# LPSTR make_sortword(LPCSTR);
# converts SJIS characters, hiragana to katakana, lower to upper for
# case-insensitive sorting.
#
sub make_sortword
{
	my(@chars) = unpack('C*', $_[0]);
	my($str, $char, $next);

	while (1) {
		$char = shift(@chars);
		last if (!defined($char) || !$char);

		if (0x82 == $char) {
			$next = shift(@chars);
			if (0x81 <= $next && $next <= 0x9a) {
				$next -= 0x21;
			} elsif (0x9f <= $next && $next <= 0xdd) {
				$char = 0x83;
				$next -= 0x5f;
			} elsif (0xde <= $next && $next <= 0xf1) {
				$char = 0x83;
				$next -= 0x5e;
			}
			$str .= pack('CC', $char, $next);

		} elsif ((0x81 <= $char && $char <= 0x9f)
				|| (0xe0 <= $char && $char <= 0xfc)) {
			$next = shift(@chars);
			if (0x81 == $char && 0x40 == $next) {
				# space in SJIS, nop.
			} else {
				$str .= pack('CC', $char, $next);
			}

		} elsif (0x2d == $char) {
			# '-', nop.

		} else {
			$str .= pack('C', $char);	# but unexpected.
		}
	}

	return($str);
}


#
# LPSTR ensure_str(LPCSTR str);
# converts sigle-byte yen to double-byte one.
#
sub ensure_str
{
	my(@chars) = unpack('C*', $_[0]);
	my($fpwobj) = $_[1];
	my($str) = '';
	my($char);

	while (0 < scalar(@chars)) {
		$char = shift(@chars);

		if ((0x81 <= $char && $char <= 0x9f)
				|| (0xe0 <= $char && $char <= 0xf1)) {
			$str .= pack('CC', $char, shift(@chars));

		} elsif (0x5c == $char) {
			#
			# yen mark.
			#
			if (defined($opt_gaiji) && defined($fpwobj)) {
				if ('' ne $str) {
					$fpwobj->add_text(Jcode::convert($str, 'euc', 'sjis')) ||
						&fatal("Failed to add text: $str", $fpwobj);
					$str = '';
				}
				$fpwobj->add_half_user_character('5c') ||
					&fatal("Failed to add gaiji for 0x5c (\'\\\')", $fpwobj);
			} else {
				$str .= "\x81\x8f";
			}

		} elsif (0xa0 <= $char && $char <= 0xdf) {
			#
			# so-called hankaku-katakana.
			#
			if (defined($opt_gaiji) && defined($fpwobj)) {
				if ('' ne $str) {
					$fpwobj->add_text(Jcode::convert($str, 'euc', 'sjis')) ||
						&fatal("Failed to add text: $str", $fpwobj);
					$str = '';
				}
				$fpwobj->add_half_user_character(sprintf("%02x", $char)) ||
					&fatal(sprintf("Failed to add gaiji for %02x (%c)",
					$char, $char), $fpwobj);
			} else {
				# FreePWING will convert it into full-shape.
				$str .= pack('C', $char);
			}

		} else {
			$str .= pack('C', $char);
		}
	}

	if (defined($fpwobj)) {
		$fpwtext->add_text(Jcode::convert($str, 'euc', 'sjis')) ||
			&fatal("Failed to add text: $str", $fpwtext);
		return('');
	}
	return($str);
}


#
# LPSTR *get_tag(LPCSTR dic, LPCSTR word);
# returns the index of the specified word.
#
sub get_tag
{
	my($word, $dic) = @_;
	my($index);

	if (!defined($dic) || '' eq $dic) {
		foreach $dic (keys %wdic_index) {
			$index = &get_tag($word, $dic);
			return($index) if ('' ne $index);
		}
	} else {
		my(@index) = @{$wdic_index{$dic}};
		$index = 0;
		while ($index < scalar(@index)) {
			if ($index[$index] eq $word) {
				return(sprintf("%s%04x", $dic, $index) . "\t$dic");
			}
			$index++;
		}
	}
	return('');
}


#
# void add_line(BOOL nowrap);
# adds one line to the FreePWING dictionary using or not using gaiji.
#
sub add_line
{
	my($nowrap, $str) = @_;

	if ('' ne $str) {
		$fpwtext->add_nowrap_start() if ($nowrap);
		$str = &ensure_str($str, $fpwtext);
		$fpwtext->add_nowrap_end() if ($nowrap);
	}
	$fpwtext->add_newline() ||
		&fatal("Failed to add a new line", $fpwtext);
}


#-----------------------------------------------------------------------#
#	appending ext plug-in												#
#-----------------------------------------------------------------------#
#
# void wdic_ext(void);
# appends 'ext.csv' plug-in if it exists in the source directory.
#
sub wdic_ext
{
	my(%ftype) = (
		BINARY => 'oCi',
		TEXT => 'eLXg'
	);
	my(%fcnt) = (
		AUDIO => 'yEȂ',
		IMAGE => '摜',
		VIDEO => '',
		COMP => 'kf[^',
		ARC => 'k𔺂ȂA[JC',
		PROG => 'vO\[Xt@C',
		EXE => 'vOst@C',
		INF => '',
		DB => 'f[^x[X',
		TEXT => ''
	);
	my($ext) = ' : gq';
	my($mime) = 'MIME ^Cv : ';
	my($file) = "$srcdir/ext.csv";
	my(@col, $text_pos, $heading_pos, $cnt, $tag);

	-f $file || return;
	open(IN, "<$file") ||
		&fatal("Failed to open plug-in file \'$file\' ($!)");
	binmode(IN);
	print "Reading plug-in file \'$file\'...\n";

	$fpwtext->new_context() ||
		&fatal("Failed to switch to the new context", $fpwtext);

	#
	# pre-convert the japanese characters to be encoded in EUC.
	#
	foreach $_ (keys %ftype) {
		$ftype{$_} = Jcode::convert($ftype{$_}, 'euc');
	}
	foreach $_ (keys %fcnt) {
		$fcnt{$_} = Jcode::convert($fcnt{$_}, 'euc');
	}
	$ext = Jcode::convert($ext, 'euc');
	$mime = Jcode::convert($mime, 'euc');

	while (<IN>) {
		s/\x0d\x0a?$//;
		@col = split(/,/);
		next if ('' eq $col[0]);

		$cnt++;
		$tag = sprintf("EXT%04x", $cnt);
		print "$tag: $col[0]\n";

		#
		# add this line as a new entry.
		#
		$fpwtext->new_entry() ||
				&fatal("Cannot start new entry", $fpwtext);
		$fpwheading->new_entry() ||
			&fatal("Cannot start new entry", $fpwheading);

		$text_pos = $fpwtext->entry_position();
		$heading_pos = $fpwheading->entry_position();

		#
		# write a tag, heading and keyword. the word is always in ASCII
		# characters in this file.
		#
		$_ = $col[0];
		$fpwheading->add_text($_) ||
			&fatal("Cannot add heading", $fpwheading);

		if (!$fpwtext->add_tag($tag)
				|| !$fpwtext->add_keyword_start()
				|| !$fpwtext->add_text($_)
				|| !$fpwtext->add_keyword_end()
				|| !$fpwtext->add_text($ext)
				|| !$fpwtext->add_newline()) {
			&fatal("Cannot add heading text", $fpwtext);
		}

		#
		# register search words.
		#
		if (!$fpwword2->add_entry($_,
				$heading_pos, $heading_file_name,
				$text_pos, $text_file_name) ||
			!$fpwword2->add_entry(".$_",
				$heading_pos, $heading_file_name,
				$text_pos, $text_file_name)) {
			&fatal("\nCannot add search word", $fpwword2);
		}

		$fpwtext->add_indent_level(2);

		#
		# file type.
		#
		if ('' ne $col[1]) {
			if (defined($ftype{$col[1]})) { $_ = $ftype{$col[1]}; }
			else { $_ = $col[1]; }
		} else {
			$_ = '';
		}
		if ('' ne $col[2]) {
			$_ .= ' / ' if ('' ne $_);
			if (defined ($fcnt{$col[2]})) { $_ .= $fcnt{$col[2]}; }
			else { $_ .= $col[2]; }
		}
		if ('' ne $_) {
			if (!$fpwtext->add_text("$opt_bullet$_")
					|| !$fpwtext->add_newline()) {
				&fatal("Failed to add text: $_", $fpwtext);
			}
		}

		#
		# MIME type.
		#
		if ('' ne $col[3]) { $_ = $col[3] }
		else { $_ = ''; }
		if ('' ne $col[4]) {
			$_ .= '/' if ('' ne $_);
			$_ .= $col[4];
		}
		if ('' ne $_) {
			if (!$fpwtext->add_text("$opt_bullet$mime$_")
					|| !$fpwtext->add_newline()) {
				&fatal("Failed to add text: $_", $fpwtext);
			}
		}

		#
		# description.
		#
		$_ = $col[5];
		if ('' ne $_) {
			if (!$fpwtext->add_text($opt_bullet)
					|| !$fpwtext->add_text(Jcode::convert($_, 'euc', 'sjis'))
					|| !$fpwtext->add_newline()) {
				&fatal("Failed to add text: $_", $fpwtext);
			}
		}
		$fpwtext->add_newline();
	}

	close(IN);
}


#-----------------------------------------------------------------------#
#	menu construction													#
#-----------------------------------------------------------------------#
#
# void wdic_menu(void)
#
sub wdic_menu
{
	my($temp, $cnt, $tag, @man, %man_toc);
	my($marker) = "__INSERT_TOC__";
	my($toc) = ('-' x 33) . '  ڎ  ' . ('-' x 33);

	print "Making menu contents...\n";
	$fpwmenu->new_context() ||
		&fatal("Failed to add a new menu context", $fpwmenu);

	#
	# write table of contents.
	#
	&write_menu_ref("MENU01", "FreePWING Łu$wdic_titlevɂ");
	&write_menu_ref("MENU02", "$wdic_title 2000 }jA");
	if (-f "$srcdir/QUESTION.TXT") {
		&write_menu_ref("MENU03", "ǎ҃AP[g");
	}

	#
	# write 'readme' page.
	#
	$temp = $0;
	$temp =~ s/[:\\\/]?[^:\\\/]+$//;
	$temp = '.' if ('' eq $temp);
	&write_menu_file("MENU01", "$temp/readme.txt");

	#
	# write the manual pages.
	#
	$temp = "$srcdir/WDIC.MAN";
	open(IN, "<$temp") ||
		&fatal("Failed to open the menu source \'$temp\' ($!)");
	binmode(IN);
	print "Writing menu text from \'$temp\'...\n";

	while (<IN>) {
		s/\x0d\x0a$//;
		if (/^\-+\s*\x81\xa1\s*($sjis_any+)\s*\x81\xa1\s*\-+$/) {
			$temp = $1;
			$cnt++;
			if ($cnt <= 1 || 23 <= $cnt) {
				push(@man, $_)
			} else {
				$tag = sprintf("MAN%02d", $cnt);
				push(@man, $marker) if (!scalar(%man_toc));
				$man_toc{$tag} = $temp;
				&jprint("$tag: $temp\n");
				&add_menu_page($tag, $_);
			}
		} else {
			if ($cnt <= 1 || 23 <= $cnt) {
				push(@man, $_);
			} else {
				&write_menu_nowrap($_);
			}
		}
	}

	close(IN);
	&add_menu_page("MENU02");

	foreach $_ (@man) {
		if ($_ eq $marker) {
			if (!$fpwmenu->add_text(Jcode::convert($toc, 'euc'))
					|| !$fpwmenu->add_newline()
					|| !$fpwmenu->add_newline()) {
				&fatal("Failed to write menu contents", $fpwmenu);
			}
			while (($tag, $temp) = each %man_toc) {
				&jprint("  $tag: $temp\n");
				$fpwmenu->add_text('    ');
				&write_menu_ref($tag, $temp);
			}
			$fpwmenu->add_newline() ||
				&fatal("Failed to add new line", $fpwmenu);
		} else {
			&write_menu_nowrap($_);
		}
	}

	#
	# write the questionary.
	#
	if (-f "$srcdir/QUESTION.TXT") {
		&write_menu_file("MENU03", "$srcdir/QUESTION.TXT");
	}
}

#
# void add_menu_page(LPCSTR tag, LPCSTR title);
# adds a new menu context with the specified tag and title.
#
sub add_menu_page
{
	my($tag, $title) = @_;

	$fpwmenu->new_context() ||
		&fatal("Failed to add a new menu context", $fpwmenu);
	$fpwmenu->add_tag($tag) ||
		&fatal("Failed to add a menu tag \'$tag\'", $fpwmenu);

	if (defined($title) && '' ne $title) {
		$fpwmenu->add_text(Jcode::convert($_, 'euc', 'sjis')) ||
			&fatal("Failed to add menu text", $fpwmenu);
	}
	$fpwmenu->add_newline() ||
		&fatal("Failed to add new line", $fpwmenu);
}

#
# void write_menu_ref(LPCSTR tag, LPCSTR title);
# adds a cross-reference to another menu entry.
#
sub write_menu_ref
{
	my($tag, $title) = @_;

	if (!$fpwmenu->add_reference_start()
			|| !$fpwmenu->add_text(
				Jcode::convert($title, 'euc'))
			|| !$fpwmenu->add_reference_end($tag)
			|| !$fpwmenu->add_newline()) {
		&fatal("Failed to write a reference to the topic \'$title'",
				$fpwmenu);
	}
}

#
# void write_menu_nowrap(LPCSTR line);
# writes one line specifying no-wrap attribute.
#
sub write_menu_nowrap
{
	my($line) = @_;

	$line = $_ if (!defined($line));
	if (defined($line) && '' ne $line) {
		$fpwmenu->add_nowrap_start();
		$fpwmenu->add_text(Jcode::convert($line, 'euc', 'sjis')) ||
			&fatal("Failed to add menu text", $fpwmenu);
		$fpwmenu->add_nowrap_end();
	}
	$fpwmenu->add_newline();
}

#
# void write_menu_file(LPCSTR tag, LPCSTR file);
# simply writes the text in the specified file into a menu entry.
#
sub write_menu_file
{
	my($tag, $file) = @_;

	open(IN, "<$file") ||
		&fatal("Failed to open the menu source \'$file\' ($!)");
	binmode(IN);
	print "Writing menu text from \'$file\'...\n";

	$_ = <IN>;
	s/\x0d\x0a$//;
	&add_menu_page($tag, $_);
	while (<IN>) {
		s/\x0d\x0a$//;
		&write_menu_nowrap($_);
	}
	close(IN);
}


#-----------------------------------------------------------------------#
#	message output														#
#-----------------------------------------------------------------------#
#
# void fatal(LPCSTR errmsg, LPVOID fpwobj);
#
sub fatal
{
	my($msg, $fpwobj) = @_;
	my(@lines) = split(/\n/, $msg);

	if (1 < scalar(@lines)) {
		print STDERR "$lines[0]\n";
		$msg = $lines[1];
	}
	if ($msg =~ /[^\x00-\x7e]/) {
		$msg = Jcode::convert($msg, $opt_charset);
	}
	if (defined($fpwobj)) {
		$msg = $msg . " (" . $fpwobj->error_message() . ')';
	}
	die "Fatal: $msg.\n";
}

#
# void warn(LPCSTR errmsg);
#
sub warn
{
	my($msg) = @_;

	if ($msg =~ /[^\x00-\x7e]/) {
		$msg = Jcode::convert($msg, $opt_charset);
	}
	print STDERR "Warning: $msg.\n";
}

#
# void jprint(LPCSTR str);
#
sub jprint
{
	my($str) = @_;

	if ($str =~ /[^\x00-\x7e]/) {
		$str = Jcode::convert($str, $opt_charset);
	}
	print $str;
}


#-----------------------------------------------------------------------#
#	usage																#
#-----------------------------------------------------------------------#
#
#	void usage(int code);
#
sub usage
{
	print<<__EOF__;
Usage: $0 [-c charset] [-b bullet] [-s sort] [-g] [inputdir]
For more information, use pod2man(1) or any similar utility.
__EOF__
	exit($_[0]);
}

__END__

=pod

=head1 NAME

wdic-fpw.pl - a FreePWING script for WDIC

=head1 SYNOPSIS

perl wdic-fpw.pl [-c charset] [-b bullet] [-s sort] [inputdir]

=head1 DESCRIPTION

This script generates a JIS X 4081 book from 'WDIC' (possibly
"WIRED & WIRELESS DICTIONARY"), available at E<lt>http://wdic.org/E<gt>, using
FreePWING. The generated book can be viewed with many EPWING viewers.

=head1 OPTIONS

=over 8

=item B<-h> | B<-?> | B<--help>

Displays this usage.

=item B<-c> I<charset> | B<--charset>=I<charset>

Specifies the display charset, which is used when printing progress. The
default is I<sjis> for Windows and I<euc> for others. This option would never
affect the book-producing routines.

=item B<-b> I<bullet> | B<--bullet>=I<bullet>

Specifies a character or string to be used as a 'bullet' for each description.
The default is a full-width white diamond.

=item B<-s> I<sort> | B<--sort>=I<sort>

Specifies the sort command you want to use. The default is I<'sort.exe /REC 65535'>
for Windows or I</usr/bin/sort> for others.

=item B<-g> | B<--gaiji>

Use half-shape user-defined characters (gaiji) for yen marks ('\') and so-called
'hankaku katakana's. If this option is not specified, they will be converted
into full-shape and no user-defined characters will be used.

=back

=head1 B<ENVIRONMENT>

=over 8

=item B<LANG>

Refered when B<--charset> option is not given.

=item B<OS> or B<COMSPEC>

Refered when B<--charset> option is not given and B<LANG> is not defined, to
distinguish the Windows systems.

=back

=head1 VERSION

$Id: wdic-fpw.pl 1.1.0.6 2000/07/16 22:53:56 satomii Exp $

=head1 AUTHOR

Rei E<lt>rei@wdic.orgE<gt>

=head1 SEE ALSO

FreePWING
E<lt>http://www.sra.co.jp/people/m-kasahr/freepwing/E<gt>

=cut
